# coding: utf8


__version__ = '0.1.19'

from .server import Server
from .connection import Connection
from .output_listener import OutputListener
from .timer import Timer
from .utils import safe_call, safe_func
from .log import logger




# coding: utf8

import time

import kcp_wrapper

from .log import logger
from .output_listener import OutputListener
from .timer import Timer


class Connection(object):

    output_listener_class = OutputListener

    # 检测过期
    expire_timer = None

    app = None
    sock = None
    address = None

    kcp_obj = None

    def __init__(self, app, sock, address):
        self.app = app
        self.sock = sock
        self.address = address

        self.expire_timer = Timer()
        self._set_expire_callback()
        self.app.events.create_conn(self)

        self.kcp_obj = self.create_kcp_obj()

    def create_kcp_obj(self):
        """
        可以继承重写
        :return:
        """
        kcp_obj = kcp_wrapper.KcpWrapper(self.app.conv)
        # kcp_obj.setNoDelay(1, 10, 2, 1)
        kcp_obj.setOutputListener(self.output_listener_class(self.sock, self.address).__disown__())
        # mtu * size，size 就是包数量
        # kcp_obj.setWndSize(100, 100)
        kcp_obj.timer = Timer()

        return kcp_obj

    def get_now_time_ms(self):
        return int(time.time() * 1000) % 0xFFFFFFFF

    def kcp_update_and_check(self):
        """
        在send和input之后调用
        :return:
        """
        now = self.get_now_time_ms()

        # update
        self.kcp_obj.update(now)

        # 调用input，需要重新check
        interval = self.kcp_obj.check(now) - now
        if interval < 0:
            interval = 0
        self.kcp_obj.timer.set(interval / 1000.0, lambda: self.kcp_obj.update(self.get_now_time_ms()), force=True)

    def write(self, data):
        if isinstance(data, self.app.box_class):
            # 打包
            data = data.pack()
        elif isinstance(data, dict):
            data = self.app.box_class(data).pack()

        self.kcp_obj.send(data)
        self.kcp_update_and_check()

    def close(self):
        """
        直接关闭连接
        """
        self.app.remove_conn(self)
        self._on_connection_close()

    def closed(self):
        """
        连接是否已经关闭
        :return:
        """
        return not self.app.has_conn(self)

    def handle(self, message):
        """
        启动处理
        """
        self.kcp_obj.input(message)
        self.kcp_update_and_check()

        while True:
            data = self.kcp_obj.recv()

            if data:
                box = self.app.box_class()
                if box.unpack(data) > 0:
                    box._raw_data = data
                    self._on_read_complete(box)
            else:
                # 没有数据了才break
                break

    def _on_connection_close(self):
        # 链接被关闭的回调

        self._clear_expire_callback()

        self.app.events.close_conn(self)

    def _on_read_complete(self, box):
        """
        数据获取结束
        data: 原始数据
        box: 解析后的box
        """

        # 每收到一次消息，就进行一次延后
        self._set_expire_callback()

        try:
            self.app.events.handle_request(self, box)
        except Exception, e:
            logger.error('view_func raise exception. box: %s, e: %s',
                         box, e, exc_info=True)

    def _set_expire_callback(self):
        """
        注册超时的回调
        :return:
        """
        if self.app.timeout:
            # 超时了，就报错
            self.expire_timer.set(self.app.timeout, self._on_timeout)

    def _clear_expire_callback(self):
        self.expire_timer.clear()

    def _on_timeout(self):
        self.app.events.timeout_conn(self)
        self.close()

    def __repr__(self):
        return '<%s address: %s>' % (self.__class__.__name__, self.address)




# coding: utf8

import gevent
from gevent.server import DatagramServer

from .mixins import AppEventsMixin
from .connection import Connection


class Server(AppEventsMixin):

    server_class = DatagramServer
    connection_class = Connection

    box_class = None

    host_list = None
    port = None
    # 连接最长不活跃时间
    timeout = None
    conv = None

    conn_dict = None

    def __init__(self, box_class, host_list, port, timeout=None, conv=None):
        super(Server, self).__init__()
        self.box_class = box_class
        self.host_list = host_list
        self.port = port
        self.timeout = timeout
        self.conv = conv or 0
        self.conn_dict = dict()

    def handle(self, sock, message, address):
        conn = self.conn_dict.get(address)
        if not conn:
            conn = self.connection_class(self, sock, address)
            self.conn_dict[address] = conn

        conn.handle(message)

    def run(self):
        job_list = []
        for host in self.host_list:
            class UDPServer(self.server_class):
                def handle(sub_self, message, address):
                    return self.handle(sub_self.socket, message, address)

            server = UDPServer((host, self.port))
            job_list.append(gevent.spawn(server.serve_forever))

        gevent.joinall(job_list)

    def remove_conn(self, conn):
        self.conn_dict.pop(conn.address, None)

    def has_conn(self, conn):
        return conn.address in self.conn_dict



# coding: utf8

import kcp_wrapper


class OutputListener(kcp_wrapper.OutputListener):
    sock = None
    address = None

    def __init__(self, sock, address):
        kcp_wrapper.OutputListener.__init__(self)
        self.sock = sock
        self.address = address

    def call(self, data, kcp_obj):
        # print 'output:', repr(data)
        offset = 0
        while offset < len(data):
            try:
                ret = self.sock.sendto(data[offset:], self.address)
            except:
                # server直接关闭，client就会抛异常
                return -1

            offset += ret

        return 0



# -*- coding: utf-8 -*-

import gevent

from .utils import safe_call


class Timer(object):

    timer = None

    def set(self, interval, callback, repeat=False, force=True):
        """
        添加timer
        """
        if self.timer:
            if force:
                # 如果已经存在，那么先要把现在的清空
                self.clear()
            else:
                # 已经存在的话，就返回了
                return

        def callback_wrapper():
            # 必须要确定，这次调用就是这个timer引起的
            if self.timer == timer:
                # 必须加这句，否则如果在callback中有clear操作，会出现GreenletExit
                self.timer = None
                # 不可以加 timer = None，否则会导致判断self.timer == timer 报错找不到timer
                result = safe_call(callback)
                if repeat and not self.timer:
                    # 之所以还要判断timer，是因为callback中可能设置了新的回调
                    self.set(interval, callback, repeat, True)
                return result

        self.timer = timer = gevent.spawn_later(interval, callback_wrapper)

    def clear(self):
        """
        直接把现在的清空
        """
        if not self.timer:
            return

        # 不阻塞
        try:
            self.timer.kill(block=False)
        except:
            pass
        self.timer = None
